/*
 * therm.c
 * 
 *   Read the temperature from a thermometer device.
 *   Sends an address byte to device and reads a temperature byte back.
 *
 *   Uses a configuration file (RCFILE) reminiscent of termcap, printcap, ...
 *   The field types are:
 *       dv=   device file                                      [no default]
 *       pa=   parity, one of 'e', 'o', 'm', 's'                [line default]
 *       br#   baud rate                                        [line default]
 *       ad#   address byte to send                             ['\n']
 *       ns#   number of samples to average (for noisy sensors) [1]
 *       is=   input format ('b' byte 'X.Yf' ascii)             [8f]
 *       to#   temperature offset (added to sensor reading)     [0]
 *       ct=   conversion, 'n' none, 'f' to F, 'c' to C         [n]
 *       fs=   output (printf) format string                    [%f\n]
 *
 *   Example rc entry:
 *   ayres|outside|outside ayres hall:\
 *       dv=/dev/tty00:br#9600:pa=none:\
 *       to#-128:ns#3:\
 *       fs=Temperature outside Ayres Hall (Knoxville TN) is %.0f degrees F\n:
 *
 *   13 jan 1995  Robert Manchek  manchek@cs.utk.edu
 *
 *   15 mar 96 updated to deal with ascii sensor outputs and C (is=, ct=)
 */

/*
 * Copyright (C) 1995 by Robert Manchek.  manchek@CS.UTK.EDU.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that this copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.
 * 
 * This software is provided "as is" without express or implied warranty.
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <signal.h>
#include <termios.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>


#define	TRIES	3		/* number of tries to read reply from device */
#define BADSAMPLE	1000000

#ifndef	RCFILE
#define	RCFILE	"/usr/local/etc/thermcap"
#endif

extern int      errno;
#ifndef __STDC__
extern char	*sys_errlist[];
#define strerror(e) sys_errlist[(e)]
#endif

static char *pname;

#ifndef __P
# if __STDC__
#  define __P(a) a
# else
#  define __P(a) ()
# endif
#endif

struct sensor {
    int   fd;
    char *ifmt;		/* sensor input format */
    char *conv;		/* type of conversion */
    char *format;	/* output format string */
    int   tof;		/* temperature offset */
    int   ns;		/* number of samples */
    int   ad;		/* address byte */
    double *ta;
    char *name;		/* name of therm to read */
    char *dev;		/* name of the device therm is connected to */
};

static void *emalloc __P((size_t));
static char *estrdup __P((char *));
static void usage __P((void));
static char **blkcpy __P((char **));
static void tformat __P((char *, size_t, char *));
static char **gettherment __P((char *));
static int breakwords __P((char *, int *, char **, char *));
static int getabyte __P((int, int, int, char *));
static char *searchent __P((char **, char *));
static int openline __P((char *, char **));
#ifdef DEBUG
static int printtio __P((struct termios *));
#endif
static int setupline __P((int, char **));
static int readsensor __P((char *, int, int, double *));
#ifdef MEDIAN
static int doublecmp __P((double *, double *));
#endif
static double getsample __P((struct sensor *));
static void dolog __P((int));


static void *
emalloc(n)
    size_t n;
{
    void *p = malloc(n);

    if (p == NULL) {
	(void) fprintf(stderr, "%s: Out of memory\n", pname);
	exit(1);
    }

    return p;
}

static char *
estrdup(s)
    char *s;
{
    size_t l = strlen(s) + 1;
    char *d = malloc(l);

    if (d == NULL) {
	(void) fprintf(stderr, "%s: Out of memory\n", pname);
	exit(1);
    }

    (void) memcpy(d, s, l);
    return d;
}
static void
usage()
{
    (void) fprintf(stderr, "Usage: %s [-l logfile] [-s seconds] [-f timefmt] [-c separator] sensor\n", pname);
}

static char **
blkcpy(av)
    char **av;
{
    char **sav = av, **dav;

    while (*av++)
	continue;

    dav = (char **) emalloc((av - sav) * sizeof(char *));

    av = sav;
    sav = dav;

    while (*av)
	*dav++ = estrdup(*av++);

    *dav = NULL;
    return sav;
}


static void
tformat(buf, len, fmt)
    char *buf;
    size_t len;
    char *fmt;
{
    time_t t;
    struct tm *tp;
    t = time(NULL);
    tp = localtime(&t);
    (void) strftime(buf, len, fmt, tp);
}

/*
 * gettherment()
 * 
 * Get a thermometer's entry from the thermcap file. Returns a null-terminated
 * list of string attributes, one for each field in the entry.
 */

static char **
gettherment(name)
    char           *name;
{
    static char    *av[20];
    static char     buf[4096];

    FILE           *ff;
    int             l;
    int             c, oc;
    int             ac;
    int             na;
    char           *nm[20];

    if ((ff = fopen(RCFILE, "r")) == NULL)
	return NULL;

    while (1) {

	/* read a device entry from the cap file */

	l = 0;
	oc = ':';
	do {
	    if ((c = getc(ff)) == EOF)
		goto notfound;
	    while (c == '\\') {
		if ((c = getc(ff)) != '\n') {
		    buf[l++] = '\\';
		    break;
		}
		c = getc(ff);
	    }
	    if (oc == ':' && (c == ' ' || c == '\t'))
		continue;
	    if (c == '\n')
		c = 0;
	    buf[l++] = c;
	    oc = c;
	} while (c);

#ifdef DEBUG
	(void) fprintf(stderr, "<%s>\n", buf);
#endif
	ac = sizeof(av) / sizeof(av[0]) - 1;
	if (breakwords(buf, &ac, av, ":") || ac < 1)
	    continue;
	av[ac] = 0;

#ifdef DEBUG
	for (l = 0; l < ac; l++)
	    (void) fprintf(stderr, "[%2d] <%s>\n", l, av[l]);
#endif

	na = sizeof(nm) / sizeof(nm[0]);
	if (breakwords(av[0], &na, nm, "|") || na < 1)
	    continue;
	if (na > 1)
	    na--;

#ifdef DEBUG
	for (l = 0; l < ac; l++)
	    (void) fprintf(stderr, "[%2d] <%s>\n", l, av[l]);
#endif

	while (--na >= 0)
	    if (!strcmp(nm[na], name))
		break;
	if (na < 0)
	    continue;

	(void) fclose(ff);
	return blkcpy(av + 1);
    }
  notfound:
    (void) fclose(ff);
    return NULL;
}


/*
 * breakwords()
 * 
 * Parse a string into words separated by whitespace. Max number of words is
 * original value of *acp.
 * 
 * Trashes out the original string.
 *
 * Returns 0 with av[0]..av[*acp - 1] pointing to the words.
 * Returns 1 if too many words.
 * Returns -1 if unbalanced quote.
 */

static int
breakwords(s, acp, av, sep)
    char           *s;		/* the string to parse */
    int            *acp;	/* max num words in, num words found out */
    char          **av;		/* pointers to words */
    char           *sep;	/* list of separator characters */
{
    int             ac = 0;	/* number of words found */
    char           *p = s;	/* input scanner */
    char           *q;		/* output */
    int             n = *acp;	/* max number of words allowed */
    int             mode = 0;	/* quote mode */

    while (*p) {
	if (ac >= n) {
	    *acp = ac;
	    return 1;
	}
	q = p;
	av[ac++] = p;
	while (*p) {
	    if (mode) {
		mode = 0;
		switch (*p) {

		case 'n':
		    *q++ = '\n';
		    break;

		case 'r':
		    *q++ = '\r';
		    break;

		case 't':
		    *q++ = '\t';
		    break;

		default:
		    *q++ = *p;
		    break;
		}

	    } else {
		if (strchr(sep, *p))
		    break;

		if (*p == '\\')
		    mode = *p;
		else
		    *q++ = *p;
	    }
	    p++;
	}
	if (*p)
	    p++;
	*q = 0;
	if (mode)
	    return -1;
    }
    *acp = ac;
    return 0;
}


/*
 * getabyte()
 * 
 * Send a byte to the thermometer device and get a response back. Flush any old
 * characters before starting. Time out in less than a second so we don't
 * hang if there's nothing there.
 */

static int
getabyte(d, ad, len, resp)
    int             d;		/* file */
    int             ad;		/* address byte to send */
    int             len;	/* length of response */
    char           *resp;	/* return response */
{
    fd_set          fds;
    struct timeval  to;
    int             t;
    int             cc;
    int             got;

    for (t = TRIES; t-- > 0;) {

	/*
	 * flush any junk waiting
	 */
	got = 1000;
	do {
	    FD_ZERO(&fds);
	    FD_SET(d, &fds);
	    to.tv_sec = 0;
	    to.tv_usec = 0;
	    if ((cc = select(d + 1, &fds, NULL, NULL, &to)) == 1)
		(void) read(d, resp, 1);
	} while ((got-- > 0 && cc == 1) || (cc == -1 && errno == EINTR));

	/*
	 * send address byte
	 */
	FD_ZERO(&fds);
	FD_SET(d, &fds);
	to.tv_sec = 0;
	to.tv_usec = 0;
	if ((cc = select(d + 1, NULL, &fds, NULL, &to)) != 1) {
	    if (cc == -1)
		(void) fprintf(stderr, "%s: Select error (%s)\n",
		    pname, strerror(errno));
	    else
		(void) fprintf(stderr, "%s: sensor port blocked\n", pname);
	    return -1;
	}
	resp[0] = ad;
	(void) write(d, resp, 1);

	/*
	 * read len bytes into resp
	 */
	for (got = 0; got < len;) {
#ifdef DEBUG
	    (void) fprintf(stderr, "len %d got %d\n", len, got);
#endif
	    FD_ZERO(&fds);
	    FD_SET(d, &fds);
	    to.tv_sec = 1;
	    to.tv_usec = 500000;
	    cc = select(d + 1, &fds, NULL, NULL, &to);
	    if (cc == -1)
		if (errno == EINTR)
		    continue;
		else {
		    (void) fprintf(stderr, "%s: Select error (%s)\n",
			pname, strerror(errno));
		    return -1;
		}
	    if (cc == 1) {
		cc = read(d, resp + got, len - got);
		if (cc == -1) {
		    (void) fprintf(stderr, "%s: Read error (%s)\n",
			pname, strerror(errno));
		    return -1;
		}
		if (cc == 0) {
		    (void) fprintf(stderr, "%s: read EOF\n", pname);
		    return -1;
		}
		got += cc;

	    } else
		break;
	}
#ifdef DEBUG
	(void) fprintf(stderr, "len %d got %d\n", len, got);
#endif
	if (got == len)
	    return 0;
    }
    return -1;
}


/*
 * searchent()
 * 
 * Search for a field in a thermcap entry and return its value.
 */

static char *
searchent(te, name)
    char **te;
    char  *name;
{
    int l = strlen(name);

    while (*te)
	if (!strncmp(*te, name, l))
	    return *te + l;
	else
	    te++;
    return NULL;
}


struct baudval {
    char           *name;
    int             speed;
    int             setting;
};

static struct baudval  rates[] =
{
    {"50", 50, B50},
    {"75", 75, B75},
    {"110", 110, B110},
    {"134", 134, B134},
    {"150", 150, B150},
    {"200", 200, B200},
    {"300", 300, B300},
    {"600", 600, B600},
    {"1200", 1200, B1200},
    {"2400", 2400, B2400},
    {"4800", 4800, B4800},
    {"9600", 9600, B9600},
#ifdef B19200
    {"19200", 19200, B19200},
#endif
#ifdef B38400
    {"38400", 38400, B38400},
#endif
#ifdef EXTA
    {"exta", -1, EXTA},
#endif
#ifdef EXTB
    {"extb", -1, EXTB},
#endif
    {0, 0}
};

/*
 * openline()
 * 
 * Open the thermometer device tty and configure it.
 */

static int
openline(dev, te)
    char  *dev;
    char **te;
{
    char *ptr;
    int   d;

    if ((ptr = strchr(dev, '@')) != NULL) {
	struct hostent *hp;
	struct sockaddr_in sa;
	unsigned long   addr;

	(void) memset(&sa, 0, sizeof(sa));
	/*
	 * Look for a host@port entry
	 */
	*ptr++ = '\0';
	if ((addr = inet_addr(dev)) == ~0) {
	    if ((hp = gethostbyname(dev)) != NULL)
		memcpy(&addr, hp->h_addr, sizeof(addr));
	    else {
		(void) fprintf(stderr, "%s: Unknown host `%s'\n", pname, dev);
		return -1;
	    }
	}
	sa.sin_family = AF_INET;
	sa.sin_port = atoi(ptr);
	sa.sin_addr.s_addr = addr;

	*--ptr = '@';

	if ((d = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
	    (void) fprintf(stderr, "%s: error opening socket (%s)\n",
		pname, strerror(errno));
	    return -1;
	}
	if (connect(d, (struct sockaddr *) & sa, sizeof(sa)) == -1) {
	    (void) fprintf(stderr, "%s: error connecting socket (%s)\n",
		pname, strerror(errno));
	    return -1;
	}
    } else {
	if ((d = open(dev, O_RDWR, 0)) == -1) {
	    (void) fprintf(stderr, "%s: error opening %s (%s)\n",
		pname, dev, strerror(errno));
	    return -1;
	}
	setupline(d, te);
    }


    return d;
}


#ifdef DEBUG
static int
printtio(tiop)
    struct termios *tiop;
{
    int             i;
    struct baudval *bp;

    (void) printf("c_iflag 0x%x, c_oflag 0x%x, c_cflag 0x%x, c_lflag 0x%x\n",
	tiop->c_iflag, tiop->c_oflag, tiop->c_cflag, tiop->c_lflag);

    i = cfgetispeed(tiop);
    for (bp = rates; bp->name; bp++)
	if (bp->setting == i)
	    break;
    if (bp->name)
	(void) printf(" input speed %s", bp->name);
    i = cfgetospeed(tiop);
    for (bp = rates; bp->name; bp++)
	if (bp->setting == i)
	    break;
    if (bp->name)
	(void) printf(" output speed %s\n", bp->name);
    (void) printf(" c_cc[");
    for (i = 0; i < NCCS; i++)
	(void) printf(" %d", (int) (tiop->c_cc[i]));
    (void) printf(" ]\n");
    return 0;
}
#endif


/*
 * setupline()
 * 
 * Configure the therm device tty.
 */

static int
setupline(d, te)
    int             d;
    char          **te;
{
    struct termios  tio;
    int             i;
    char           *br;
    char           *pa;
    struct baudval *bp;

    if (tcgetattr(d, &tio) == -1) {
	if (errno != ENOTTY && errno != EINVAL)
	    (void) fprintf(stderr, "%s: error getting terminal attributes (%s)\n",
		pname, strerror(errno));
	return 1;
    }
#ifdef DEBUG
    printtio(&tio);
#endif

    tio.c_iflag = IGNPAR | INPCK;
    /*
     * tio.c_oflag = ONL0 | CR0 | TAB0 | BS0 | VT0 | FF0;
     */
    tio.c_oflag = 0;
#ifdef	CBAUD
    tio.c_cflag = (tio.c_cflag & CBAUD) | CREAD | CLOCAL | CS8;
#else
    tio.c_cflag = CREAD | CLOCAL | CS8;
#endif
    tio.c_lflag = 0;
    tio.c_cc[VMIN] = 0;
    tio.c_cc[VTIME] = 0;

    if ((pa = searchent(te, "pa=")) != NULL) {
	switch (*pa) {

	case 'e':
	    tio.c_cflag |= PARENB;
	    break;

	case 'o':
	    tio.c_cflag |= (PARODD | PARENB);
	    break;

#ifdef DEBUG
	case 'm':
	    tio.c_cflag |= (PARODD|PAREXT|PARENB);
	    break;

	case 's':
	    tio.c_cflag |= (PAREXT|PARENB);
	    break;
#endif

	default:
	    break;
	}
    }
    if ((br = searchent(te, "br#")) != NULL) {
	i = atoi(br);
	for (bp = rates; bp->name; bp++)
	    if (bp->speed == i)
		break;
	if (bp->name) {
	    cfsetospeed(&tio, (speed_t) bp->setting);
	    cfsetispeed(&tio, (speed_t) bp->setting);
	} else {
	    (void) fprintf(stderr, "%s: unknown baud rate: %d\n", pname, i);
	}
    }
    if (tcsetattr(d, TCSAFLUSH, &tio) == -1)
	(void) fprintf(stderr, "%s: error setting terminal attributes (%s)\n",
		pname, strerror(errno));

    if (tcgetattr(d, &tio) == -1)
	(void) fprintf(stderr, "%s: error getting terminal attributes (%s)\n",
		pname, strerror(errno));

#ifdef DEBUG
    printtio(&tio);
#endif
    return 0;
}


/*
 * readsensor()
 * 
 * Get a reading from the sensor.
 */

static int
readsensor(ifmt, fd, ad, r)
    char           *ifmt;	/* sensor output format */
    int             fd;		/* file */
    int             ad;		/* address of sensor */
    double         *r;		/* reading return */
{
    int             w1 = 0;	/* field width */
    int             w2 = 0;	/* precision */
    char            buf[40];

    w1 = 1;
    if (isdigit(*ifmt)) {
	w1 = atoi(ifmt);
	while (isdigit(*++ifmt));
    }
    w2 = 0;
    if (*ifmt == '.') {
	ifmt++;
	if (isdigit(*ifmt)) {
	    w2 = atoi(ifmt);
	    while (isdigit(*++ifmt));
	}
    }
    switch (*ifmt) {
    case 'b':
	if (getabyte(fd, ad, 1, buf) < 0)
	    goto noresp;
#ifdef DEBUG
	(void) fprintf(stderr, "0x%x\n", buf[0] & 0xff);
#endif
	*r = buf[0] & 0xff;
	break;

    case 'f':
	if (getabyte(fd, ad, w1, buf) < 0)
	    goto noresp;
	buf[w1] = 0;
	if (sscanf(buf, "%lf", r) != 1) {
	    (void) fprintf(stderr, "%s: no sensor\n", pname);
	    *r = 0;
	    return -1;
	}
	break;

    default:
	(void) fprintf(stderr, "%s: unknown sensor format: '%c'\n", 
	    pname, *ifmt);
	return -1;
    }
    return 0;

  noresp:
    (void) fprintf(stderr, "%s: no response from device\n", pname);
    return -1;
}


#ifdef	MEDIAN
static int
doublecmp(a, b)
    double *a, *b;
{
    if (*a > *b)
	return 1;
    else if (*a < *b)
	return -1;
    return 0;
}
#endif


static double
getsample(se)
	struct sensor *se;
{
    double  t, ts;
    int i;
    
    ts = 0;
    for (i = se->ns; i-- > 0;) {
	if (readsensor(se->ifmt, se->fd, se->ad, &t) < 0)
	    return BADSAMPLE;
	t += se->tof;
#ifdef DEBUG
	(void) fprintf(stderr, "%f F\n", t);
#endif
	se->ta[i] = t;
	ts += t;
    }

#ifdef	MEDIAN
    qsort(se->ta, se->ns, sizeof(double), doublecmp);
    ts = se->ta[se->ns / 2];
#else
    ts /= se->ns;
#endif

    switch (se->conv[0]) {
    case 'f':
	ts = (ts * 9) / 5 + 32;
	break;
    case 'c':
	ts = ((ts - 32) * 5) / 9;
	break;
    default:
	break;
    }
    return ts;
}


static int	   recycle = 0;

static void
dolog(n)
    int n;
{
    recycle = 1;
}

int
main(argc, argv)
    int             argc;
    char          **argv;
{
    int             i;
    char          **te;
    char           *p;
    struct sensor  *se, *cse, *bse;
    extern int      optind;
    extern char    *optarg;
    char	   *sep = "\t";
    int		    sec = 30;
    char	   *fmt = "%H%M%S";
    char	   *logfile = NULL;
    FILE	   *logfp = NULL;

    if ((pname = strrchr(argv[0], '/')) != NULL)
	pname++;
    else
	pname = argv[0];

    while ((i = getopt(argc, argv, "l:c:s:f:")) != -1)
	switch (i) {
	case 'f':
	    fmt = optarg;
	    break;
	    
	case 's':
	    sec = atoi(optarg);
	    break;

	case 'l':
	    logfile = optarg;
	    break;

	case 'c':
	    sep = optarg;
	    break;

	default:
	    usage();
	    break;
	}

    bse = cse = (struct sensor *) emalloc((argc - optind) *
					  sizeof(struct sensor));

    for (i = optind; i < argc; i++) {
	cse->ifmt = "8f";
	cse->conv = "n";
	cse->format = "%f\n";
	cse->tof = 0;
	cse->ns = 1;
	cse->ad = 0;
	cse->name = argv[i];

	if (!(te = gettherment(cse->name))) {
	    (void) fprintf(stderr, "%s: unknown device: %s\n", pname,
		cse->name);
	    continue;
	}
	if (!(cse->dev = searchent(te, "dv="))) {
	    (void) fprintf(stderr, "%s: bad configuration for %s\n", pname,
		cse->name);
	    continue;
	}
	if ((p = searchent(te, "is=")) != NULL)
	    cse->ifmt = p;
	if ((p = searchent(te, "ct=")) != NULL)
	    cse->conv = p;
	if ((p = searchent(te, "fs=")) != NULL)
	    cse->format = p;
	if ((p = searchent(te, "to#")) != NULL)
	    cse->tof = atoi(p);
	if ((p = searchent(te, "ns#")) != NULL) {
	    cse->ns = atoi(p);
	    if (cse->ns < 1)
		cse->ns = 1;
	    else if (cse->ns > 1000)
		cse->ns = 1000;
	}
	if ((p = searchent(te, "ad#")) != NULL)
	    cse->ad = atoi(p);

	cse->ta = (double *) emalloc(cse->ns * sizeof(double));

	for (se = bse; se < cse; se++)
	    if (strcmp(se->dev, cse->dev) == 0) {
		cse->fd = se->fd;
		break;
	    }

	if (se == cse)
	    if ((cse->fd = openline(cse->dev, te)) == -1) {
		free(cse->ta);
		continue;
	    }

	cse++;
    }

    if (logfile) {
	struct sigaction sa;

	sa.sa_handler = dolog;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = 0;
	sigaction(SIGHUP, &sa, NULL);
	recycle = 1;
    }

    do {
	char buf[1024];

	if (recycle) {
	    if (logfp) {
		fclose(logfp);
		logfp = NULL;
	    }
	    tformat(buf, sizeof(buf), logfile);
	    logfp = fopen(buf, "w");

	    if (logfp != NULL) {
		(void) fprintf(logfp, "Time");
		for (se = bse; se < cse; se++)
		    (void) fprintf(logfp, "%s%s", sep, se->name);
		(void) fprintf(logfp, "\n");
	    }

	    recycle = 0;
	}


	if (logfp != NULL) {
	    tformat(buf, sizeof(buf), fmt);
	    (void) fprintf(logfp, "%s", buf);
	}
	for (se = bse; se < cse; se++) {

	    double  ts = getsample(se);
	    if (logfp) {
		if (ts == BADSAMPLE)
		    (void) fprintf(logfp, "%s?", sep);
		else
		    (void) fprintf(logfp, "%s%g", sep, ts);
	    }
	    else {
		if (ts == BADSAMPLE)
		    continue;
		(void) printf(se->format, ts);
	    }
	}
	if (logfp) {
	    struct timeval tv;
	    (void) fprintf(logfp, "\n");
	    (void) fflush(logfp);
	    tv.tv_sec = sec;
	    tv.tv_usec = 0;
	    (void) select(0, NULL, NULL, NULL, &tv);
	}
    }
    while (logfp);
    return 0;
}
