/*
 *	Serial.c
 *
 *	Allow one to talk to a serial port.
 *
 *	usage: serial [-78x] [-b rate] [device]
 *
 *	Mar 1989  Robert Manchek  manchek@boulder.Colorado.EDU.
 *	Apr 1990  Added baudrate ifdefs to make use of whatever is available.
 *	Mar 1991  Added decent cmd processor with room for expansion.
 *	          Added hexdump for incoming chars.
 *	May 1992  Added send command (may hang, cause it doesn't read from tty).
 *	Oct 1995  Converted to termios, added command line options
 *	          I hate termios.  filth filth filth.
 *	$Log: serial.c,v $
 *	Revision 1.3  1998/08/27 14:18:09  manchek
 *	added some more baud rates
 *
 *	Revision 1.2  1998/08/27 14:12:37  manchek
 *	some ANSI tweaks
 *
 */

#include <sys/param.h>
#include <sys/types.h>
#ifdef	NEEDSYSSELECTH
#include <sys/select.h>
#endif
#include <fcntl.h>
#include <termios.h>
#include <ctype.h>
#include <stdio.h>
#ifdef	SYSVSTR
#include <string.h>
#define	CINDEX(s,c)	strchr(s,c)
#else
#include <strings.h>
#define	CINDEX(s,c)	index(s,c)
#endif

#ifndef FD_SET

#ifndef NBBY
#define NBBY 8
#endif /*NBBY*/

#ifndef FD_SETSIZE
#define FD_SETSIZE (sizeof(fd_set) * 8)
#endif /*FD_SETSIZE*/

#define FD_SET(n,p) ((p)->fds_bits[(n)/8] |= (1 << ((n) & 7)))

#ifndef FD_CLR
#define FD_CLR(n,p) ((p)->fds_bits[(n)/8] &= ~(1 << ((n) & 7)))
#endif /*FD_CLR*/

#ifndef FD_ISSET
#define FD_ISSET(n,p) ((p)->fds_bits[(n)/8] & (1 << ((n) & 7)))
#endif /*FD_ISSET*/

#ifndef FD_ZERO
#define FD_ZERO(p) bzero((char*)p, sizeof(*(p)))
#endif /*FD_ZERO*/

#endif /*FD_SET*/

#define	LNEXTC	('V' - 64)
#define	ESCAPEC	(']' - 64)
#define MAXARGS 10		/* for commands */

struct baudval {
	char *name;
	int speed;
	int setting;
};

struct baudval rates[] = {
	{ "50", 50, B50 },
	{ "75", 75, B75 },
	{ "110", 110, B110 },
	{ "134", 134, B134 },
	{ "150", 150, B150 },
	{ "200", 200, B200 },
	{ "300", 300, B300 },
	{ "600", 600, B600 },
	{ "1200", 1200, B1200 },
	{ "2400", 2400, B2400 },
	{ "4800", 4800, B4800 },
	{ "9600", 9600, B9600 },
#ifdef B19200
	{ "19200", 19200, B19200 },
#endif
#ifdef B38400
	{ "38400", 38400, B38400 },
#endif
#ifdef B57600
	{ "57600", 57600, B57600 },
#endif
#ifdef B115200
	{ "115200", 115200, B115200 },
#endif
#ifdef B230400
	{ "230400", 230400, B230400 },
#endif
#ifdef B460800
	{ "460800", 460800, B460800 },
#endif
#ifdef EXTA
	{ "exta", -1, EXTA },
#endif
#ifdef EXTB
	{ "extb", -1, EXTB },
#endif
	{ 0, 0 }
};


struct termios origtio, tlktio;		/* for saving control tty modes */
char devnam[256];					/* for tty filename */
int port = -1;						/* tty fd */
int hexmode = 0;					/* display incoming in hex */
int bit8mode = 1;					/* 8 bit passthrough */
int mylnextchr = LNEXTC;

int setspeed();
char *getspeed();

static char rcsid[] = "$Id: serial.c,v 1.3 1998/08/27 14:18:09 manchek Exp $";

int
main(argc, argv)
	int argc;
	char **argv;
{
	int i, j;
	int err = 0;
	char *p;
	char *baud = 0;

	for (i = j = 1; i < argc; i++) {
		if (argv[i][0] == '-')
			for (p = argv[i] + 1; *p; p++) {
				switch (*p) {
				case 'b':
					if (++i == argc) {
						fprintf(stderr, "missing parameter for -b\n");
						err++;

					} else {
						baud = argv[i];
					}
					break;

				case '7':
					bit8mode = 0;
					break;

				case '8':
					bit8mode = 1;
					break;

				case 'x':
					hexmode = 1;
					break;

				default:
					fprintf(stderr, "unknown switch: %c\n", *p);
					err++;
					break;
				}
			}

		else
			argv[j++] = argv[i];
	}
	argc = j;

	if (err || argc > 2)
		goto usage;

	(void)strcpy(devnam, "no device");
	if (argc == 2) {
		openline(argv[1]);
		if (baud)
			setspeed(port, baud, 3);
/*
		printf("%s speed %s\n", argv[1], getspeed(port, 2));
		pukemode(port);
*/
	}

	if (!isatty(0)) {
		fputs("Standard input must be a tty\n", stderr);
		exit(2);
	}

/*
	pukemode(0);
*/
	if (tcgetattr(0, &origtio)) {
		perror("tcgetattr");
		exit(2);
	}
	tlktio = origtio;
	tlktio.c_iflag &= ~(ICRNL);
#ifdef OCRNL
	tlktio.c_oflag &= ~(ONLCR|OCRNL);
#else
	tlktio.c_oflag &= ~(ONLCR);
#endif
	tlktio.c_lflag &= ~(ISIG|ICANON|ECHO);
/*
	tlktio.c_cc[VINTR] = -1;
	tlktio.c_cc[VQUIT] = -1;
	tlktio.c_cc[VSUSP] = -1;
	tlktio.c_cc[VDSUSP] = -1;
*/
	tlktio.c_cc[VLNEXT] = -1;
	tlktio.c_cc[VMIN] = 0;
	tlktio.c_cc[VTIME] = 0;
	mylnextchr = origtio.c_cc[VLNEXT];
	
	/* command loop */

	cmd("serial>");
	exit(0);

usage:
	fputs("usage: serial [-78x] [-b rate] [devname]\n", stderr);
	exit(1);
}


/*********************
*  support routines  *
*                    *
*********************/


int
pukemode(port)
	int port;
{
	struct termios tio;
	char buf[256];

	if (tcgetattr(port, &tio))
		perror("tcgetattr");

	buf[0] = 0;
	if (tio.c_iflag & IGNBRK) strcat(buf, " IGNBRK");
	if (tio.c_iflag & BRKINT) strcat(buf, " BRKINT");
	if (tio.c_iflag & IGNPAR) strcat(buf, " IGNPAR");
	if (tio.c_iflag & PARMRK) strcat(buf, " PARMRK");
	if (tio.c_iflag & INPCK) strcat(buf, " INPCK");
	if (tio.c_iflag & ISTRIP) strcat(buf, " ISTRIP");
	if (tio.c_iflag & INLCR) strcat(buf, " INLCR");
	if (tio.c_iflag & IGNCR) strcat(buf, " IGNCR");
	if (tio.c_iflag & ICRNL) strcat(buf, " ICRNL");
#ifdef IUCLC
	if (tio.c_iflag & IUCLC) strcat(buf, " IUCLC");
#endif
	if (tio.c_iflag & IXON) strcat(buf, " IXON");
	if (tio.c_iflag & IXANY) strcat(buf, " IXANY");
	if (tio.c_iflag & IXOFF) strcat(buf, " IXOFF");
	if (tio.c_iflag & IMAXBEL) strcat(buf, " IMAXBEL");
	fprintf(stderr, "c_iflag = 0x%x %s\n", tio.c_iflag, buf);

	buf[0] = 0;
	if (tio.c_oflag & OPOST) strcat(buf, " OPOST");
#ifdef OLCUC
	if (tio.c_oflag & OLCUC) strcat(buf, " OLCUC");
#endif
	if (tio.c_oflag & ONLCR) strcat(buf, " ONLCR");
#ifdef OCRNL
	if (tio.c_oflag & OCRNL) strcat(buf, " OCRNL");
#endif
#ifdef ONOCR
	if (tio.c_oflag & ONOCR) strcat(buf, " ONOCR");
#endif
#ifdef ONLRET
	if (tio.c_oflag & ONLRET) strcat(buf, " ONLRET");
#endif
#ifdef OFILL
	if (tio.c_oflag & OFILL) strcat(buf, " OFILL");
#endif
#ifdef OFDEL
	if (tio.c_oflag & OFDEL) strcat(buf, " OFDEL");
#endif
	fprintf(stderr, "c_oflag = 0x%x %s\n", tio.c_oflag, buf);

	buf[0] = 0;
	if (tio.c_lflag & ISIG) strcat(buf, " ISIG");
	if (tio.c_lflag & ICANON) strcat(buf, " ICANON");
#ifdef XCASE
	if (tio.c_lflag & XCASE) strcat(buf, " XCASE");
#endif
	if (tio.c_lflag & ECHO) strcat(buf, " ECHO");
	if (tio.c_lflag & ECHOE) strcat(buf, " ECHOE");
	if (tio.c_lflag & ECHOK) strcat(buf, " ECHOK");
	if (tio.c_lflag & ECHONL) strcat(buf, " ECHONL");
	if (tio.c_lflag & NOFLSH) strcat(buf, " NOFLSH");
	if (tio.c_lflag & TOSTOP) strcat(buf, " TOSTOP");
	if (tio.c_lflag & ECHOCTL) strcat(buf, " ECHOCTL");
	if (tio.c_lflag & ECHOPRT) strcat(buf, " ECHOPRT");
	if (tio.c_lflag & ECHOKE) strcat(buf, " ECHOKE");
	if (tio.c_lflag & FLUSHO) strcat(buf, " FLUSHO");
	if (tio.c_lflag & PENDIN) strcat(buf, " PENDIN");
	if (tio.c_lflag & IEXTEN) strcat(buf, " IEXTEN");
	fprintf(stderr, "c_lflag = 0x%x %s\n", tio.c_lflag, buf);

	buf[0] = 0;
	fprintf(stderr, "c_cflag = 0x%x %s\n", tio.c_cflag, buf);

	return 0;
}


int
talk(port, hexmode)
	int port;
	int hexmode;
{
	fd_set rds;
	char buf[512];
	int n;
	int col = 0;
	int i, j;
	int lnex = 0;
	int wantout = 0;

	if (port < 0) {
		fputs("Not Connected.\n", stderr);
		return 0;
	}

	if (tcsetattr(0, TCSAFLUSH, &tlktio))
		perror("tcsetattr");

	/* character exchange loop */

	while (1) {
		FD_ZERO(&rds);
		FD_SET(0, &rds);
		FD_SET(port, &rds);
/*
		fprintf(stderr, "into select\n");
*/
		if (select(port + 1, &rds, 0, 0, 0) == -1) {
			perror("select()");
			break;
		}
/*
		fprintf(stderr, "outof select\n");
*/
		if (FD_ISSET(0, &rds)) {
/*
			fprintf(stderr, "read 0\n");
*/
			n = read(0, buf, sizeof(buf));
			if (n < 0) {
				perror("read(0)");
				break;
			}
			if (n > 0) {
				for (i = j = 0; i < n; i++) {
					if (!lnex) {
						if (buf[i] == mylnextchr) {
							lnex = 1;
							continue;
						} else if (buf[i] == ESCAPEC) {
							wantout = 1;
							break;
						}
					}
					buf[j++] = buf[i];
					lnex = 0;
				}
/*
				fprintf(stderr, "%d bytes to port\n", j);
				hdump(stderr, buf, j);
*/
				if (j > 0)
					write(port, buf, j);
				if (wantout)
					break;
			}
		}
		if (FD_ISSET(port, &rds)) {
/*
			fprintf(stderr, "read port\n");
*/
			n = read(port, buf, sizeof(buf));
			if (n < 0) {
				perror("read(port)");
				break;
			}
/*
			fprintf(stderr, "%d bytes from port\n", n);
			hdump(stderr, buf, n);
*/
			if (n > 0) {
				if (!bit8mode)
					for (i = n; i-- > 0; )
						buf[i] &= 0x7f;
				if (hexmode)
					for (i = 0; i < n; i++) {
						fprintf(stderr, "%02x%s",
							0xff & buf[i], (col == 15 ? "\r\n" : " "));
						col = (col + 1) & 0xf;
					}
				else
					write(1, buf, n);
			}
		}
	}
	if (tcsetattr(0, TCSAFLUSH, &origtio))
		perror("tcsetattr");
	fputc('\n', stderr);
	return 0;
}


int
openline(name)
	char *name;
{
	struct termios tio;

	if (port != -1)
		close(port);
	port = -1;

	if (CINDEX(name, '/'))
		devnam[0] = 0;
	else
		(void)strcpy(devnam, "/dev/");
	(void)strncat(devnam, name, sizeof(devnam)-6);

	if ((port = open(devnam, O_RDWR|O_NONBLOCK, 0)) < 0) {
		perror(devnam);
	} else {
		if (isatty(port)) {
			if (tcgetattr(port, &tio))
				perror("tcgetattr");
			tio.c_iflag = 0;
			tio.c_oflag = 0;
#ifdef	CBAUD
			tio.c_cflag = (tio.c_cflag & CBAUD) | CREAD | CLOCAL | CS8;
#else
			tio.c_cflag = CREAD | CLOCAL | CS8;
#endif
			tio.c_lflag = 0;
			tio.c_cc[VMIN] = 0;
			tio.c_cc[VTIME] = 0;
			if (tcsetattr(port, TCSAFLUSH, &tio))
				perror("tcsetattr");
		}
	}
	return port;
}


/*
*	setspeed()
*
*	Set baud rate of tty given int baud rate.
*
*	how =
*		1 set input rate
*		2 set output rate
*		3 set both rates
*
*/

int
setspeed(line, speed, how)
	int line;
	char* speed;
	int how;
{
	struct termios tio;
	struct baudval *bp;

	for (bp = rates; bp->name; bp++)
		if (!strcmp(bp->name, speed))
			break;
	if (!bp->name)
		return -1;

	if (tcgetattr(line, &tio))
		perror("tcgetattr");
	if (how & 1)
		cfsetispeed(&tio, (speed_t)bp->setting);
	if (how & 2)
		cfsetospeed(&tio, (speed_t)bp->setting);
	if (tcsetattr(line, TCSAFLUSH, &tio))
		perror("tcsetattr");
	return 0;
}


/*	getspeed()
*
*	Get baud rate of tty
*/

char*
getspeed(line, how)
	int line;
	int how;
{
	struct termios tio;
	struct baudval *bp;
	int i, o;

	if (tcgetattr(line, &tio)) {
		perror("tcgetattr");
	} else {
		o = cfgetospeed(&tio);
		i = cfgetispeed(&tio);
		for (bp = rates; bp->name; bp++)
			if (bp->setting == ((how & 2) ? o : i))
				return bp->name;
	}
	return "unknown";
}


int
truestring(s)
	char *s;
{
	return (!strcmp(s, "1")
		|| !strcmp(s, "y")
		|| !strcmp(s, "yes")
		|| !strcmp(s, "on")
		|| !strcmp(s, "true"));
}


/*	hdump()
*
*	Hex dump of byte array.
*/

void
hdump(ff, p, n)
	FILE *ff;
	char *p;	/* bytes */
	int n;		/* length */
{
	int i;
	for (i = 0; i < n; i++)
		fprintf(ff, "%02x%s", 0xff & (int)p[i],
			(i < n - 1) ? (((i & 15) == 15) ? "\n" : " ") : "\n");
}


/*********************
*  command routines  *
*                    *
*********************/

int baud_cmd();
int bit8_cmd();
int help_cmd();
int hex_cmd();
int quit_cmd();
int send_cmd();
int tty_cmd();

/* command table */

struct cmdsw {
	char *cmd;		/* official name */
	int a1;			/* min arg cnt */
	int a2;			/* max arg cnt */
	int (*fun)();	/* function */
};

static struct cmdsw commands[] = {
	{ "baud", 1, 2, baud_cmd },
	{ "bit8", 1, 2, bit8_cmd },
	{ "help", 1, 2, help_cmd },
	{ "hex", 1, 2, hex_cmd },
	{ "quit", 1, 1, quit_cmd },
	{ "send", 2, 2, send_cmd },
	{ "tty", 1, 2, tty_cmd },
	{ 0, 0, 0, 0 }
};


int
baud_cmd(ac, av)
	int ac;
	char **av;
{
	struct baudval *bp;

	if (ac == 1) {
		if (port == -1)
			fprintf(stderr, "not connected\n");
		else
			fprintf(stderr, "speed set to %s\n", getspeed(port, 2));
	} else {
		if (av[1][0] == '?' && !av[1][1]) {
			fputs("Available settings:\n", stderr);
			for (bp = rates; bp->name; bp++)
				fprintf(stderr, "%s ", bp->name);
			fputc('\n', stderr);
			return 0;
		}
		if (!setspeed(port, av[1], 3)) {
			fprintf(stderr, "speed set to %s\n", getspeed(port, 2));
		}
		else
			fprintf(stderr, "can't set speed to %s\n", av[1]);
	}
	return 0;
}


int
tty_cmd(ac, av)
	int ac;
	char **av;
{
	if (ac > 1) {
		openline(av[1]);

	} else {	/* just asking */
		fprintf(stderr, "%s\n", devnam);
	}
	return 0;
}


int
hex_cmd(ac, av)
	int ac;
	char **av;
{
	if (ac > 1)
		hexmode = truestring(av[1]);
	else
		hexmode = !hexmode;
	fprintf(stderr, "hexmode is %s\n", hexmode ? "on" : "off");
	return 0;
}


int
bit8_cmd(ac, av)
	int ac;
	char **av;
{
	if (ac > 1)
		bit8mode = truestring(av[1]);
	else
		bit8mode = !bit8mode;
	fprintf(stderr, "bit8 passthrough is %s\n", bit8mode ? "on" : "off");
	return 0;
}


int
quit_cmd(ac, av)
	int ac;
	char **av;
{
	exit(0);
}


int
send_cmd(ac, av)
	int ac;
	char **av;
{
	int d, n;
	char buf[4096];

	if (port < 0) {
		fputs("Not Connected.\n", stderr);
		return 0;
	}
	if ((d = open(av[1], O_RDONLY, 0)) == -1) {
		perror(av[1]);
		return 0;
	}
	while ((n = read(d, buf, sizeof(buf))) > 0)
		write(port, buf, n);
	(void)close(d);
	return 0;
}

/* command aliases */

static char *aliases[] = {
	"8", "bit8",
	"?", "help",
	"b", "baud",
	"h", "help",
	"q", "quit",
	"t", "tty",
	"x", "hex",
	0
};

/* help cmd output by topic */

static char *helptx[] = {
	"help help - Give helpful info",
	"help Syntax: help [topic]  or  h [topic]  or  ? [topic]",
	"help Commands are:",
	"baud baud - Set/show tty speed",
	"baud Syntax: baud [rate]  or  b [rate]",
	"baud   ? lists available settings",
	"bit8 bit8 - Toggle 8 bit passthrough",
	"bit8 Syntax: bit8 [how]  or 8 [how]",
	"quit quit - Terminate program",
	"quit Syntax: quit  or  q",
	"send send - Send file to serial port",
	"send Syntax: send filename",
	"tty tty - Set/Show tty port name",
	"tty Syntax: tty [device]  or  t [device]",
	"hex hex - Toggle hex output mode",
	"hex Syntax: hex [how]  or  x [how]",
	0
};


int
help_cmd(ac, av)
	int ac;
	char **av;
{
	char **p;
	char *topic;
	int l;
	struct cmdsw *csp;

	/* if not specified, topic = help */
	if (ac > 1)
		topic = av[1];
	else
		topic = "help";

	for (p = aliases; *p; p += 2)
		if (!strcmp(topic, *p)) {
			topic = *++p;
			break;
		}
	l = strlen(topic);

	/* search through messages for ones matching topic */
	for (p = helptx; *p; p++) {
		if (!strncmp(topic, *p, l) && ((*p)[l] == ' ' || (*p)[l] == '-'))
			fprintf(stderr, "%s\n", (*p) + l + 1);
	}

	if (!strcmp(topic, "help")) {
		for (csp = commands; csp->cmd; csp++) {
			l = strlen(csp->cmd);
			for (p = helptx; *p; p++)
				if (!strncmp(csp->cmd, *p, l) && (*p)[l] == ' ') {
					fprintf(stderr, "%s\n", (*p) + l + 1);
					break;
				}
		}
	}
	return 0;
}


int
cmd(prompt)
	char *prompt;
{
	int ac;
	char *av[MAXARGS];
	struct cmdsw *csp;
	char **p;
	char buf[256];
	int n;

	while (1) {
		if (prompt && *prompt)
			fputs(prompt, stderr);
		if ((n = read(0, buf, sizeof(buf))) < 1) {
			fputs("quit\n", stderr);
			exit(0);
		}
		buf[n] = 0;
		ac = sizeof(av)/sizeof(av[0]);
		if (acav(buf, &ac, av)) {
			fputs("command too long\n", stderr);
			continue;
		}
		if (!ac) {
			talk(port, hexmode);
			continue;
		}
		for (p = aliases; *p; p += 2)
			if (!strcmp(av[0], *p)) {
				p++;
				break;
			}
		if (!*p)
			p = &av[0];
		for (csp = commands; csp->cmd; csp++) {
			if (!strcmp(csp->cmd, *p)) {
				if (ac >= csp->a1 && ac <= csp->a2)
					(csp->fun)(ac, av);
				else
					fputs("incorrect arg count\n", stderr);
				break;
			}
		}
		if (!csp->cmd)
			fputs("unknown command\n", stderr);
	}
}


/*	acav()
*
*	Parse a string into words separated by whitespace.
*	Max number of words is original value of *acp.
*
*	Trashes out the original string.
*	Returns 0 with av[0]..av[*acp - 1] pointing to the words.
*	Returns 1 if too many words.
*/

int
acav(s, acp, av)
	char *s;		/* the string to parse */
	int *acp;		/* max words in, ac out */
	char **av;		/* pointers to args */
{
	register int ac;
	register char *p = s;
	register n = *acp;

	/* separate out words of command */

	ac = 0;
	while (*p) {
		while (isspace(*p)) p++;
		if (*p) {
			if (ac >= n) {
	/* command too long */
				*acp = ac;
				return 1;
			}
			av[ac++] = p;
			while (*p && !isspace(*p)) p++;
			if (*p) *p++ = 0;
		}
	}
	*acp = ac;
	return 0;
}


