
/*
 *	$Id: imalloc.c,v 1.1 1996/11/17 03:21:41 manchek Exp $
 *
 *	Instrumented malloc filter.
 *
 *	7 Jun 1991  Robert Manchek  manchek@CS.UTK.EDU.
 *	6 Sep 1991  added static glob and check features.
 *	17 Dec 1991 added i_realloc, cleaned up.
 *	8 Sep 1992  added total byte count.
 *	30 Oct 1992 added object type tag.
 *	5 Mar 1993  replaced the tag with Keith's studly __FILE__ macros.
 *	26 Mar 1993 added tags to i_free and i_realloc()
 *	23 Mar 1995 added i_bitch to complain about calling raw malloc funcs.
 *	16 Nov 1995 added macro to allow freeing null pointer.
 *
 *	To use, include imalloc.h in your code and use macros:
 *		MYALLOC(n, g)		to alloc n bytes, tagged with string g
 *		MYREALLOC(p, n)		to realloc p as n bytes
 *		MYFREE(p)			to free p
 *		TALLOC(n, t, g)		to alloc n of type t, tagged with string g
 *		TREALLOC(p, n, t)	to realloc p as n of type t
 *		STRALLOC(s)			to alloc and copy a string s
 *
 *	Or just include imalloc_retro.h in your code and hope for the best.
 *	recompile, and link.
 *
 *	Features:
 *
 *	$ All errors written to fd 2 (not using stdio).  This can be
 *		changed by defining SWRITE macro.
 *
 *	$ Configurable to die on error for debugger.
 *		Define DIEONERROR as 0 or 1.
 *
 *	$ Configurable to use statically allocated space for recordkeeping.
 *		This makes imalloc more immune to really bad heap space trashing.
 *		Define STATICGLOBS as nonzero.
 *
 *	$ Configurable to allow zero-length alloc (as it's pretty useful).
 *
 *	$ i_malloc():
 *		Error on [adjustable] unreasonble length or malloc() failed.
 *		Hashes descriptors of all chunks for i_free() checking.
 *		Numbers requests to indicate order received.
 *		Writes [adjustable] pseudo-random head and tail pads to
 *			aid in checking overrun of chunk.
 *
 *	$ i_free():
 *		Error if chunk not i_malloc()d.
 *		Checks head and tail pads.
 *		[Optionally] zeros chunk to aid in detecting later use.
 *		Define ZEROONFREE as 0 or 1.
 *
 *	$ i_realloc():
 *		Performs checks of both i_free and i_malloc.
 *
 *	$ i_dump(how):
 *		Callable from user program.
 *		Dumps the hash table to see what chunks are active, each record
 *		Shows:
 *			Chunk id (Serial number)
 *			'*' if chunk was allocd since last i_dump()
 *			'+' if chunk was reallocd since last i_dump()
 *			File and line where i_malloc was called
 *			The chunk address and length
 *			Ascii dump of first 16 bytes of chunk
 *		If how is 1, i_dump() checks each chunk's head and tail pads.
 */

/*
*	$Log: imalloc.c,v $
*	Revision 1.1  1996/11/17 03:21:41  manchek
*	Initial revision
*
*	Revision 1.2  1996/11/12 06:21:28  manchek
*	code cleanup
*
*	Revision 1.1  1996/11/12 05:19:19  manchek
*	Initial revision
*
*/

#include <stdlib.h>
#include <stdio.h>
#ifdef	SYSVSTR
#include <string.h>
#else
#include <strings.h>
#endif

#ifndef	BZERO
#define	BZERO(p,n)	bzero(p,n)
#endif

/******************
*  customization  *
*                 *
******************/

#ifndef	DIEONERROR
#define	DIEONERROR	1
#endif
#ifndef	ZEROONFREE
#define	ZEROONFREE	1
#endif
#ifndef	STATICGLOBS
#define	STATICGLOBS	0
#endif
#ifndef	LET0BE1
#define	LET0BE1	1
#endif
#ifndef	LETFREENULL
#define	LETFREENULL	0
#endif
#ifndef	DEFMAXCHUNK
#define	DEFMAXCHUNK 1000000000
#endif
#ifndef	DUMPHDR
#define	DUMPHDR	"i_dump()"
#endif

/* log a null-term string */
#ifndef	SWRITE
#if 0
#define	SWRITE(fd,s)	log_error(s)
#endif
#define	SWRITE(fd,s)	write((fd),(s),strlen(s))
#endif

/*
*	end of customization
*/

#define	NEXTRN(x)	(x = (x) + (x) + (((x ^ (x >> 3)) & 0x2000) ? 1 : 0))
#define	HASH(p)		(((long)(p) ^ ((long)(p) >> 8) ^ ((long)(p) >> 16) ^ ((long)(p) >> 24)) & 0xff)

/* this describes a chunk of memory */

struct glob {
	struct glob *next;	/* next glob in hash bucket */
	char *base;			/* baseaddr of user chunk */
	int len;			/* len of user chunk */
	int id;				/* chunk id */
	int lop;			/* nbytes head padding */
	int hip;			/* nbytes tail padding */
	int rst;			/* starting random state */
	int flg;
	int itag;			/* content tag */
	char ctag[16];
};

#define	OBALLOC		1	/* ob was just alloced - cleared by dump() */
#define	OBREALLOC	2	/* ob was just realloced - cleared by dump() */

/* default values */

static int debfd = 2;				/* debug file descriptor */
static int lengthlimit = DEFMAXCHUNK;	/* max length malloc allowed */
static int lopad = 16;				/* chunk head pad */
static int hipad = 16;				/* chunk tail pad */

/* globals */

static int totlnbyts = 0;			/* total bytes allocated */
static int rnstate = 1;				/* random sequence gen. */
static struct glob *hashtbl[256];	/* chunk hash table */
static char msbuf[256];				/* error message buffer */
static int firsttime = 1;
static int globid = 0;				/* chunk id counter */
#if	STATICGLOBS > 0
static struct glob globheap[STATICGLOBS];
static struct glob *globfl = 0;
static int globavail = STATICGLOBS;
#endif
static char rcsid[] = "$Id: imalloc.c,v 1.1 1996/11/17 03:21:41 manchek Exp $";


/*	i_choke()
*
*	Found a problem - bail out if DIEONERROR is true.
*/

int
i_choke(ctag, itag)
	char *ctag;
	int itag;
{
	(void)sprintf(msbuf, "Called from %s:%d\n", ctag, itag);
	(void)SWRITE(debfd, msbuf);
#if	DIEONERROR
	abort();
#endif
}


/*	i_malloc()
*
*	Allocate a buffer of given length.
*/

char*
i_malloc(len, itag, ctag)
	unsigned len;	/* number of bytes */
	int itag;		/* content description */
	char *ctag;
{
	char *ptr;
	struct glob *ob;	/* hash tbl entry */
	struct glob **he;
	int i;				/* gp */

	if (firsttime) {
		firsttime = 0;
		BZERO((char*)hashtbl, sizeof(hashtbl));

#if	STATICGLOBS > 0
	/* initialize the glob freelist */
		ob = 0;
		for (i = STATICGLOBS-1; i >= 0; i--) {
			globheap[i].next = ob;
			ob = &globheap[i];
		}
		globfl = ob;
#endif
	}

	/* check req length */

#if LET0BE1 > 0
	if (!len)
		len = 1;
#endif
	if (len < 1 || len > lengthlimit) {
		(void)sprintf(msbuf, "i_malloc: bogus length=%d\n", len);
		(void)SWRITE(debfd, msbuf);
		i_choke(ctag, itag);
		return (char*)0;
	}

	/* do actual malloc */

	if (!(ptr = (char*)malloc(len + lopad + hipad))) {
		(void)sprintf(msbuf, "i_malloc: malloc failed, length=%d\n", len);
		(void)SWRITE(debfd, msbuf);
		i_choke(ctag, itag);
		return (char*)0;
	}

	/* get descriptor */

#if STATICGLOBS > 0
	if (ob = globfl) {
		globfl = globfl->next;
		globavail--;

	} else {
		(void)sprintf(msbuf, "i_malloc: out of globs (static limit %d)\n",
			STATICGLOBS);
		(void)SWRITE(debfd, msbuf);
		i_choke(ctag, itag);
		return (char*)0;
	}
#else
	if (!(ob = (struct glob*)malloc(sizeof(struct glob)))) {
		(void)sprintf(msbuf, "i_malloc: malloc failed for glob\n");
		(void)SWRITE(debfd, msbuf);
		i_choke(ctag, itag);
		return (char*)0;
	}
#endif

	/* enter descriptor, write head and tail pads */

	ob->flg = OBALLOC;
	ob->id = ++globid;
	ob->itag = itag;
	ob->ctag[0] = 0;
	if (ctag)
		strncpy(ob->ctag, ctag, 16);
	ob->len = len;
	ob->lop = lopad;
	ob->hip = hipad;
	ob->rst = rnstate;
	for (i = lopad; i-- > 0; *ptr++ = NEXTRN(rnstate));
	ob->base = ptr;
	he = &hashtbl[HASH(ptr)];
	for (i = hipad, ptr += len; i-- > 0; *ptr++ = NEXTRN(rnstate));
	ob->next = *he;
	*he = ob;
	totlnbyts += len;
	return ob->base;
}


/*	i_free()
*
*	Give up a previously allocated buffer.
*/

int
i_free(loc, itag, ctag)
	char *loc;		/* ptr to buffer */
	int itag;		/* location */
	char *ctag;
{
	struct glob *ob;	/* freeing this object */
	struct glob *preob;	/* object before in chain */
	int rs;				/* reproduced random seqn */
	char *ptr = loc;	/* gp */
	struct glob **he;	/* hash tbl entry */
	int i;				/* gp */

#if	LETFREENULL
	if (loc == 0)
		return 0;
#endif

	/* sanity check */

	if (firsttime) {
		char *s = "i_free: I was called before i_malloc?\n";
		SWRITE(debfd, s);
		i_choke(ctag, itag);
		return 0;
	}

	/* delete from hash tbl */

	he = &hashtbl[HASH(loc)];
	for (preob = 0, ob = *he; ob && ob->base != loc; preob = ob, ob = ob->next);
	if (!ob) {
		(void)sprintf(msbuf, "i_free: bogus pointer=0x%lx\n", (unsigned long)loc);
		(void)SWRITE(debfd, msbuf);
/*
i_dump(0);
*/
		i_choke(ctag, itag);
		return 0;
	}
	rs = ob->rst;

	/* check head and tail pads */

	for (i = ob->lop, ptr -= i; i > 0; i--)
		if ((0xff & (int)(*ptr++)) != (0xff & NEXTRN(rs))) {
			(void)sprintf(msbuf, "i_free: scribbled in 0x%lx[%d]\n", (unsigned long)loc, -i);
			(void)SWRITE(debfd, msbuf);
			i_choke(ctag, itag);
		}
	for (i = ob->hip, ptr += ob->len; i > 0; i--)
		if ((0xff & (int)(*ptr++)) != (0xff & NEXTRN(rs))) {
			(void)sprintf(msbuf, "i_free: scribbled in 0x%lx[%d+%d]\n",
				(unsigned long)loc, ob->len, ob->hip - i);
			(void)SWRITE(debfd, msbuf);
			i_choke(ctag, itag);
		}

	/* do actual free */

#if	ZEROONFREE
	BZERO(loc - ob->lop, ob->len + ob->lop + ob->hip);
#endif
	free(loc - ob->lop);
	totlnbyts -= ob->len;

	/* reclaim descriptor */

	if (preob)
		preob->next = ob->next;
	else
		*he = ob->next;
#if STATICGLOBS > 0
	ob->next = globfl;
	globfl = ob;
	globavail++;
#else
	free((char*)ob);
#endif
	return 0;
}


/*	i_realloc()
*
*	Resize a previously allocated buffer (possibly relocate as well)
*	and leave the contents unchanged up to the minimum of the old
*	and new lengths.
*/

char*
i_realloc(loc, len, itag, ctag)
	char *loc;			/* old buffer */
	unsigned len;		/* length of new buffer */
	int itag;			/* location */
	char *ctag;
{
	struct glob *ob;	/* freeing this object */
	struct glob *preob;	/* object before in chain */
	int rs;				/* reproduced random seqn */
	char *ptr = loc;	/* gp */
	struct glob **he;	/* hash tbl entry */
	int i;				/* gp */

#if	LETFREENULL
	if (loc == 0)
		return i_malloc(len, itag, ctag);
#endif

	/* sanity check */

	if (firsttime) {
		char *s = "i_realloc: I was called before i_malloc?\n";
		SWRITE(debfd, s);
		i_choke(ctag, itag);
		return (char*)0;
	}

	/* check req length */

#if LET0BE1 > 0
	if (!len)
		len = 1;
#endif
	if (len < 1 || len > lengthlimit) {
		(void)sprintf(msbuf, "i_realloc: bogus length=%d\n", len);
		(void)SWRITE(debfd, msbuf);
		i_choke(ctag, itag);
		return (char*)0;
	}

	/* delete from hash tbl */

	he = &hashtbl[HASH(loc)];
	for (preob = 0, ob = *he; ob && ob->base != loc; preob = ob, ob = ob->next);
	if (!ob) {
		(void)sprintf(msbuf, "i_realloc: bogus pointer=0x%lx\n", (unsigned long)loc);
		(void)SWRITE(debfd, msbuf);
		i_choke(ctag, itag);
		return (char*)0;
	}
	rs = ob->rst;

	/* check head and tail pads */

	for (i = ob->lop, ptr -= i; i > 0; i--)
		if ((0xff & (int)(*ptr++)) != (0xff & NEXTRN(rs))) {
			(void)sprintf(msbuf, "i_realloc: scribbled in 0x%lx[%d]\n", (unsigned long)loc, -i);
			(void)SWRITE(debfd, msbuf);
			i_choke(ctag, itag);
		}
	for (i = ob->hip, ptr += ob->len; i > 0; i--)
		if ((0xff & (int)(*ptr++)) != (0xff & NEXTRN(rs))) {
			(void)sprintf(msbuf, "i_realloc: scribbled in 0x%lx[%d+%d]\n",
				(unsigned long)loc, ob->len, ob->hip - i);
			(void)SWRITE(debfd, msbuf);
			i_choke(ctag, itag);
		}

	/* remove descriptor */

	if (preob)
		preob->next = ob->next;
	else
		*he = ob->next;

	/* realloc */

	if (!(ptr = (char*)realloc(loc - ob->lop, len + lopad + hipad))) {
		(void)sprintf(msbuf, "i_realloc: malloc failed, length=%d\n", len);
		(void)SWRITE(debfd, msbuf);
		i_choke(ctag, itag);
		return (char*)0;
	}

	/* rewrite descriptor, write head and tail pads */

	totlnbyts += len - ob->len;
	ob->flg = OBREALLOC;
	ob->id = ++globid;	/* XXX ? */
	ob->len = len;
	ob->lop = lopad;
	ob->hip = hipad;
	ob->rst = rnstate;
	for (i = lopad; i-- > 0; *ptr++ = NEXTRN(rnstate));
	ob->base = ptr;
	he = &hashtbl[HASH(ptr)];
	for (i = hipad, ptr += len; i-- > 0; *ptr++ = NEXTRN(rnstate));
	ob->next = *he;
	*he = ob;
	return ob->base;
}


/*	ascdump()
*
*	Convert byte string to printable characters.
*/

static void
ascdump(o, p, e)
	char **o;		/* addr of pointer to output space (passed back) */
	char *p;		/* string to convert */
	int e;			/* length of string */
{
	char *r = *o;
	char c;

	while (e-- > 0) {
		c = 0x7f & *p;
		if (c < ' ' || c == 0x7f) {
			c = (c + '@') & 0x7f;
			*r++ = '^';
		} else
			*r++ = ' ';
		*r++ = c;
		p++;
	}
	*r++ = '\n';
	*r = 0;
	*o = r;
}


/*	i_dump()
*
*	Dump the table of contents of allocated buffers.
*/

int
i_dump(how)
	int how;			/* != 0 also do sanity checking */
{
	int ht;				/* hash table index */
	struct glob *ob;
	char *r;
	int i;
	int rs;				/* reproduced random seqn */
	char *ptr;			/* gp */
	int err;
	char ctagstr[17];

#if	STATICGLOBS > 0
	sprintf(msbuf, "%s %d bytes total/%d globs free\n",
			DUMPHDR, totlnbyts, globavail);
	(void)SWRITE(debfd, msbuf);
#else
	sprintf(msbuf, "%s %d bytes total\n", DUMPHDR, totlnbyts);
	(void)SWRITE(debfd, msbuf);
#endif
	for (ht = 0; ht < 256; ht++) {
		for (ob = hashtbl[ht]; ob; ob = ob->next) {
			rs = ob->rst;
			ptr = ob->base;
			err = 0;
			if (how) {
				for (i = ob->lop, ptr -= i; i > 0; i--)
					if ((0xff & (int)(*ptr++)) != (0xff & NEXTRN(rs))) {
						(void)sprintf(msbuf,
							"%5d 0x%08lx[%d]: scribbled in [%d]\n",
							ob->id, (unsigned long)ob->base, ob->len, -i);
						(void)SWRITE(debfd, msbuf);
						err++;
					}
				for (i = ob->hip, ptr += ob->len; i > 0; i--)
					if ((0xff & (int)(*ptr++)) != (0xff & NEXTRN(rs))) {
						(void)sprintf(msbuf,
							"%5d 0x%08lx[%d]: scribbled in [%d+%d]\n",
							ob->id, (unsigned long)ob->base, ob->len, ob->len, ob->hip - i);
						(void)SWRITE(debfd, msbuf);
						err++;
					}
			}
			if (!err) {
				strncpy(ctagstr, ob->ctag, 16);
				ctagstr[16] = 0;
				sprintf(msbuf, "%5d%c%16s#%4d 0x%08lx[%4d]",
					ob->id,
					(ob->flg & OBALLOC ? '*' :
						(ob->flg & OBREALLOC ? '+' : ' ')),
					ctagstr,
					ob->itag,
					(unsigned long)ob->base, ob->len);
				r = msbuf + strlen(msbuf);
				*r++ = ' ';
				if ((i = ob->len) > 24)
					i = 24;
				ascdump(&r, ob->base, i);
				(void)SWRITE(debfd, msbuf);
			}
			ob->flg &= ~(OBALLOC|OBREALLOC);
		}
	}
	return 0;
}


/*	i_test()
*
*	Returns 0 if pointer was not returned by i_malloc or i_realloc.
*/

int
i_test(loc)
	char *loc;			/* ptr to buffer */
{
	struct glob *ob;

	/* sanity check */

	if (firsttime)
		return 0;

	/* delete from hash tbl */

	for (ob = hashtbl[HASH(loc)]; ob && ob->base != loc; ob = ob->next) ;
	return ob ? 1 : 0;
}


int
i_bitch(s, l, f)
	char *s;
	int l;
	char *f;
{
	sprintf(msbuf, "imalloc: aargh, you called %s()\n", s);
	(void)SWRITE(debfd, msbuf);
	i_choke(f, l);
	return 0;
}


